// 002 u_c_calcul_charges_sarl
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_charges_sarl;
  interface
    uses Classes, u_c_basic_object
        , u_strings, u_stringlist, u_c_stringlist

        , u_c_parameters
        ;

    type c_calcul_charges_sarl=
             class(c_basic_object)
               m_c_parameters_ref: c_parameters;
               m_c_detail_result_list: tStrings;
               m_title: String;

               m_month_count: Integer;

               m_c_key_list, m_c_value_list: tStringList;

               m_total_charges_salaire_sarl, m_pourcent_charges_salariales, m_percent_dividendes: Double;
               m_csg_deductible_sarl: Double;

               Constructor create_calcul_charges_sarl(p_name: String; p_c_parameters_ref: c_parameters);

               procedure load_sarl_parameters(p_path, p_file_name: String);
               function f_amount(p_key: String): Double;
               function f_rate(p_key: String): Double;
               procedure display_sarl_parameters;

               procedure compute_charges_sociales_sarl(p_salaire, p_cotisations_sociales_obligatoire_precedentes: Double);

               Destructor Destroy; Override;
             end; // c_calcul_charges_sarl

  implementation
    uses SysUtils, Math, u_c_display
        , u_utils
        ;

    // -- c_calcul_charges_sarl

    Constructor c_calcul_charges_sarl.create_calcul_charges_sarl(p_name: String; p_c_parameters_ref: c_parameters);
      begin
        Inherited create_basic_object(p_name);
        m_c_parameters_ref:= p_c_parameters_ref;

        m_title:= 'CALCUL CHARGES SALARIALES SARL';
        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;

        m_month_count:= 1;
      end; // create_calcul_charges_sarl

    procedure c_calcul_charges_sarl.load_sarl_parameters(p_path, p_file_name: String);
      var l_c_stringlist: c_stringlist;

      procedure _add_key_value(p_key, p_value: String);
        begin
          display('add '+ p_key+ ' '+ p_value);
          m_c_key_list.Add(p_key);
          m_c_value_list.Add(p_value);
        end; // _add_key_value

      var l_key, l_value: String;

      begin // load_parameters
        l_c_stringlist:= c_stringlist.create_stringlist('', p_path, p_file_name);

        with l_c_stringlist do
        begin
          if m_c_stringlist.Count= 0
            then begin
                m_c_parameters_ref.add_error(p_file_name+ ' SARL_parameter_file_empty');
                Free;
                Exit;
             end;

          while not f_end_of_file do
          begin
            if not f_is_blank_or_comment
              then begin
                  l_key:= f_next_value; l_value:= f_next_value;
                  _add_key_value(l_key, l_value);
                end
              else
                if m_title= ''
                  then m_title:= m_trimmed_line
                  else _add_key_value('','');

            read_line;
          end; // while not f_end_of_file
        end; // with l_c_stringlist

        display(m_c_key_list.Text);
        display(m_c_value_list.Text);
      end; // load_sarl_parameters

    function c_calcul_charges_sarl.f_amount(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_amount

    function c_calcul_charges_sarl.f_rate(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_rate

    procedure c_calcul_charges_sarl.display_sarl_parameters;
      var l_sarl_parameter_index: Integer;
          l_key: String;
      begin
        display_line;
        display(m_title);

        with m_c_key_list do
          for l_sarl_parameter_index:= 0 to Count- 1 do
          begin
            l_key:= Strings[l_sarl_parameter_index];
            if l_key= ''
              then display_line
              else display(Format('%-25s %10.2f', [l_key, f_amount(l_key)]));
          end;
      end; // display_sarl_parameters

    procedure c_calcul_charges_sarl.compute_charges_sociales_sarl(p_salaire, p_cotisations_sociales_obligatoire_precedentes: Double);
      var l_salaire_et_dividendes_soumis_aux_charges: Double;

      procedure _display_and_add_charge(p_name, p_threshold: String;
          p_base, p_rate, p_value, p_category_total: Double);
        var l_category_total: String;
        begin
          p_value:= p_value/ m_month_count;

          if p_category_total= 0
            then l_category_total:= ''
            else l_category_total:= Format('  %10.2f', [p_category_total]);

          m_total_charges_salaire_sarl:= m_total_charges_salaire_sarl+ p_value;
          display(Format('%-31s %-14s %10s %5.2f %10s %s ',
              [p_name, p_threshold, f_10_2(p_base), p_rate* 100, f_10_2(p_value), l_category_total]));
        end; // _display_and_add_charge

      procedure _compute_maladie;
        var l_value, l_rate: Double;
        begin
          if l_salaire_et_dividendes_soumis_aux_charges< 0.40* f_amount('plafond_secu_sarl')
            then begin
                l_rate:= (f_rate('maladie_T1')- f_rate('maladie_T2'))/ (1.10* f_amount('plafond_secu_sarl'))* l_salaire_et_dividendes_soumis_aux_charges
                    + (f_rate('maladie_T2')- f_rate('maladie_T3'))/ (0.40* f_amount('plafond_secu_sarl'))* l_salaire_et_dividendes_soumis_aux_charges
                    + f_rate('maladie_T3');
                l_rate:= l_rate/ 100;
                l_value:= l_salaire_et_dividendes_soumis_aux_charges* l_rate;
                _display_and_add_charge('maladie_1', '< 0.4* pss', l_salaire_et_dividendes_soumis_aux_charges, l_rate, l_value, 0);
              end
            else
              if l_salaire_et_dividendes_soumis_aux_charges< 1.10* f_amount('plafond_secu_sarl')
                then begin
                    l_rate:= (f_rate('maladie_T1')- f_rate('maladie_T2'))/ (1.10* f_amount('plafond_secu_sarl'))* l_salaire_et_dividendes_soumis_aux_charges
                        + f_rate('maladie_T2') ;
                    l_rate:= l_rate/ 100;
                    l_value:= l_salaire_et_dividendes_soumis_aux_charges* l_rate;
                    _display_and_add_charge('maladie_1', '< 1.1* pss', l_salaire_et_dividendes_soumis_aux_charges, l_rate, l_value, 0);
                  end
                else begin
                    l_rate:= f_rate('maladie_T1');
                    l_rate:= l_rate/ 100;
                    l_value:= l_salaire_et_dividendes_soumis_aux_charges* l_rate;
                    _display_and_add_charge('maladie_1', '> 1.1* pss', l_salaire_et_dividendes_soumis_aux_charges, l_rate, l_value, 0);
                  end;
        end; // _compute_maladie

      procedure _compute_allocations;
        var l_value, l_rate: Double;
        begin
          if l_salaire_et_dividendes_soumis_aux_charges< 1.10* f_amount('plafond_secu_sarl')
            then begin
                display('no_alloc');
              end
            else
              if (l_salaire_et_dividendes_soumis_aux_charges>= 1.10* f_amount('plafond_secu_sarl')) and (l_salaire_et_dividendes_soumis_aux_charges< 1.40* f_amount('plafond_secu_sarl'))
                then begin
                    l_rate:= f_rate('allocations_T1')/ (0.30* f_amount('plafond_secu_sarl'))*
                        (l_salaire_et_dividendes_soumis_aux_charges- 1.10*  f_amount('plafond_secu_sarl')) ;
                    l_rate:= l_rate/ 100;
                    l_value:= l_salaire_et_dividendes_soumis_aux_charges* l_rate;
                    _display_and_add_charge('alloc_2', '1.1 1.4 pss', l_salaire_et_dividendes_soumis_aux_charges, l_rate, l_value, 0);
                  end
                else
                  if (l_salaire_et_dividendes_soumis_aux_charges>= 1.40* f_amount('plafond_secu_sarl'))
                    then begin
                        l_rate:= f_rate('allocations_T1')/ 100;
                        l_value:= l_salaire_et_dividendes_soumis_aux_charges* l_rate;
                        _display_and_add_charge('alloc_3', '> 1.4 pss', l_salaire_et_dividendes_soumis_aux_charges, l_rate, l_value, 0);
                      end;
        end; // _compute_allocations

      procedure _compute_retraite_base;
        var l_base, l_rate, l_value: Double;
        begin
          l_base:= Min(l_salaire_et_dividendes_soumis_aux_charges, f_amount('plafond_secu_sarl'));
          l_rate:= f_rate('retraite_base_1')/ 100;
          l_value:= l_base* l_rate;
          _display_and_add_charge('retraite_1', '< pss', l_base, l_rate, l_value, 0);

          l_base:= l_salaire_et_dividendes_soumis_aux_charges- f_amount('plafond_secu_sarl');
          if l_base> 0
            then begin
                l_rate:= f_rate('retraite_base_2')/ 100;
                l_value:= l_base* l_rate;
                _display_and_add_charge('retraite_2', '> pss', l_base, l_rate, l_value, 0);
              end;
        end; // _compute_retraite_base

      procedure _compute_retraite_complementaire;
        var l_base, l_rate, l_value: Double;
        begin
          l_base:= Min(l_salaire_et_dividendes_soumis_aux_charges, f_amount('plafond_complementaire'));
          l_rate:= f_rate('retraite_complementaire_1')/ 100;
          l_value:= l_base* l_rate;
          _display_and_add_charge('complem_1', '< compl', l_base, l_rate, l_value, 0);

          l_base:= l_salaire_et_dividendes_soumis_aux_charges- f_amount('plafond_complementaire');
          if l_base> 0
            then begin
                l_base:= Min(l_base, 4* f_amount('plafond_secu_sarl'));
                l_rate:= f_rate('retraite_complementaire_2')/ 100;
                l_value:= l_base* l_rate;
                _display_and_add_charge('complem_2', '> compl', l_base, l_rate, l_value, 0);
              end
        end; // _compute_retraite_complementaire

      procedure _compute_formation_professionnelle;
        var l_base, l_rate, l_value: Double;
        begin
          l_base:= f_amount('plafond_secu_sarl');

          l_rate:= f_rate('formation_professionnelle_sarl')/ 100;
          l_value:= l_base* l_rate;
          _display_and_add_charge('form_profess', 'plaf', l_base, l_rate, l_value, 0);
        end; // _compute_formation_professionnelle

      var l_charges_dividendes_non_soumis: Double;

      procedure _compute_csg_crds;
        var l_base, l_rate, l_value: Double;
        begin
          l_base:= l_salaire_et_dividendes_soumis_aux_charges+ p_cotisations_sociales_obligatoire_precedentes;
          l_rate:= f_rate('csg_crds_non_deductible')/ 100;
          l_value:= l_base* l_rate;
          _display_and_add_charge('csg_crds', 'sal+cot_p', l_base, l_rate, l_value, 0);

          l_rate:= f_rate('csg_deductible')/ 100;
          l_value:= l_base* l_rate;
          _display_and_add_charge('csg_deduc', 'sal+cot_p', l_base, l_rate, l_value, 0);
          m_csg_deductible_sarl:= l_value;
        end; // _compute_csg_crds

      var l_c_save_display_strings: tStrings;

      begin // compute_charges_sociales_sarl
        with m_c_parameters_ref do
        begin
          l_c_save_display_strings:= s_c_strings;
          initialize_display(m_c_detail_result_list);
          clear_display;

          display_line;
          display('----- Calcul de charges Sarl');

          display(f_display('cotisations n-1',            p_cotisations_sociales_obligatoire_precedentes));
          display_line;

          l_salaire_et_dividendes_soumis_aux_charges:= p_salaire;

          display(f_display('salaires',                   p_salaire));

          m_total_charges_salaire_sarl:= 0;
          l_charges_dividendes_non_soumis:= 0;

          _compute_maladie;

          display_line;
          _compute_allocations;
          display_line;
          _compute_retraite_base;
          _compute_retraite_complementaire;
          display_line;

          display(Format('total_cotisations_obligatoires %s %10s', [f_spaces(43), f_10_2(m_total_charges_salaire_sarl)]));

          display_line;
          _compute_formation_professionnelle;

          display_line;
          _compute_csg_crds;

          m_percent_dividendes:= 0;

          if p_salaire= 0
            then m_pourcent_charges_salariales:= 0
            else m_pourcent_charges_salariales:= m_total_charges_salaire_sarl/ p_salaire;


          display_line;
          display(Format('total_charges %s %10s   %6.2f',
              [f_spaces(58+ 5- 14), f_10_2(m_total_charges_salaire_sarl), m_pourcent_charges_salariales]));

          // -- approx
          if m_dividendes> 0
            then begin
                // -- taux sur ce qui est soumis
                if p_salaire= 0
                  then m_pourcent_charges_salariales:= 0
                  else m_pourcent_charges_salariales:=  (m_total_charges_salaire_sarl- l_charges_dividendes_non_soumis)/ p_salaire;

              end
            else m_percent_dividendes:= 0  ;
        end; // with m_c_parameters_ref
        initialize_display(l_c_save_display_strings);
        display_line;
      end; // compute_charges_sociales_sarl

    Destructor c_calcul_charges_sarl.Destroy;
      begin
        m_c_key_list.Free;
        m_c_value_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_calcul_charges_sarl
    end. // u_c_calcul_charges_sarl



