// 002 u_c_calcul_charges_sas
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_charges_sas;
  interface
    uses Classes, u_c_basic_object
        , u_c_parameters

        , u_c_sas_parameter_list_2
        ;

    type c_calcul_charges_sas=
             class(c_basic_object)
               m_c_parameters_ref: c_parameters;
               m_c_detail_result_list: tStrings;
               m_title: String;

               m_c_sas_parameter_list_2_ref: c_sas_parameter_list_2;
               m_month_count: Integer;

               m_total_charges_salaire_sas, m_pourcent_charges_salariales: Double;
               m_csg_deductible_sas: Double;

               Constructor create_calcul_charges_sas(p_name: String; p_c_parameters_ref: c_parameters;
                   p_c_sas_parameter_list_2_ref: c_sas_parameter_list_2);
               procedure compute_charges_sociales_sas(p_salary: Double);
               Destructor Destroy; Override;
             end; // c_calcul_charges_sas

  implementation
    uses SysUtils, Math
        , u_c_display
        , u_utils
        ;

    // -- c_calcul_charges_sas

    Constructor c_calcul_charges_sas.create_calcul_charges_sas(p_name: String;
        p_c_parameters_ref: c_parameters;
        p_c_sas_parameter_list_2_ref: c_sas_parameter_list_2);
      begin
        Inherited create_basic_object(p_name);
        m_c_parameters_ref:= p_c_parameters_ref;

        m_title:= 'CHARGES SALARIALES SAS';

        m_c_sas_parameter_list_2_ref:= p_c_sas_parameter_list_2_ref;
      end; // create_calcul_charges_sas

    procedure c_calcul_charges_sas.compute_charges_sociales_sas(p_salary: Double);
      var l_total_employee, l_total_company: Double;
          l_plafond_secu: Double;
          l_prevoyance_pour_csg: Double;

      procedure _display_and_add(p_key, p_policy: String;
          p_base, p_employee_rate, p_employee_amout, p_company_rate, p_company_amount: Double);

        procedure _format(p_rate, p_value: Double; var pv_rate, pv_value: String);
          begin
            pv_rate:= ''; pv_value:= '';
            if p_rate<> 0
              then begin
                  pv_rate:= Format('%7.3f', [p_rate]);
                  pv_value:= f_10_2(p_value);
                end;
          end; // _format

        var l_employee_rate, l_employee_amount: String;
            l_company_rate, l_company_amount: String;

        begin // _display_and_add
          if (p_employee_rate= 0) and (p_company_rate= 0)
            then Exit;

          _format(p_employee_rate, p_employee_amout, l_employee_rate, l_employee_amount);
          _format(p_company_rate, p_company_amount, l_company_rate, l_company_amount);

          display(Format('%-27s %-5s %10s   %7s %10s     %7s %10s',
               [p_key, p_policy, f_10_2(p_base), l_employee_rate, l_employee_amount, l_company_rate, l_company_amount]));

          l_total_employee:= l_total_employee+ p_employee_amout;
          l_total_company:= l_total_company+ p_company_amount;
          if p_key= 'prevoyance_cadre'
            then l_prevoyance_pour_csg:= l_prevoyance_pour_csg+ p_company_amount;
        end; // _display_and_add

      procedure _compute_on_full_salary(p_c_sas_parameter: c_sas_parameter);
        var l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_employee_amount:= p_salary* m_employee_rate/ 100;
            l_company_amount:= p_salary* m_company_rate/ 100;
            _display_and_add(m_name, m_policy, p_salary,
                m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
          end;
        end; // _compute_on_full_salary

      procedure _compute_on_plafond_secu(p_c_sas_parameter: c_sas_parameter);
        var l_base: Double;
            l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(l_plafond_secu, p_salary);
            l_employee_amount:= l_base* m_employee_rate/ 100;
            l_company_amount:= l_base* m_company_rate/ 100;
            _display_and_add(m_name, m_policy, l_base,
                m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
          end;
        end; // _compute_on_plafond_secu

      procedure _compute_retraite_1_a_8_plafond(p_c_sas_parameter: c_sas_parameter);
        var l_base: Double;
            l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(l_plafond_secu* 8, p_salary);
            l_base:= l_base- l_plafond_secu;
            if l_base> 0
              then begin
                  l_employee_amount:= l_base* m_employee_rate/ 100;
                  l_company_amount:= l_base* m_company_rate/ 100;
                  _display_and_add(m_name, m_policy, l_base,
                      m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
                end;
          end;
        end; // _compute_retraite_1_a_8_plafond

      procedure _compute_on_total_up_to_4_plafond(p_c_sas_parameter: c_sas_parameter);
        var l_base: Double;
            l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(p_salary, l_plafond_secu* 4);
            l_employee_amount:= l_base* m_employee_rate/ 100;
            l_company_amount:= l_base* m_company_rate/ 100;
            _display_and_add(m_name, m_policy, l_base,
                m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
          end;
        end; // _compute_on_total_up_to_4_plafond

      procedure _compute_on_total_up_to_8_plafond(p_c_sas_parameter: c_sas_parameter);
        var l_base: Double;
            l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(p_salary, l_plafond_secu* 8);
            l_employee_amount:= l_base* m_employee_rate/ 100;
            l_company_amount:= l_base* m_company_rate/ 100;
            _display_and_add(m_name, m_policy, l_base,
                m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
          end;
        end; // _compute_on_total_up_to_8_plafond

      procedure _compute_on_total_1_to_8_plafond(p_c_sas_parameter: c_sas_parameter);
        var l_base: Double;
            l_employee_amount, l_company_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(p_salary, l_plafond_secu* 8)- l_plafond_secu;
            if l_base> 0
              then begin
                  l_employee_amount:= l_base* m_employee_rate/ 100;
                  l_company_amount:= l_base* m_company_rate/ 100;
                  _display_and_add(m_name, m_policy, l_base,
                      m_employee_rate, l_employee_amount, m_company_rate, l_company_amount);
                end;
          end;
        end; // _compute_on_total_1_to_8_plafond

      const k_deductible= True;
            k_not_deductible= False;

      procedure _compute_csg(p_c_sas_parameter: c_sas_parameter; p_is_deductible: Boolean);
        // SAS: csg ddictobme ?
        var l_base: Double;
            l_employee_amount: Double;
        begin
          with p_c_sas_parameter do
          begin
            l_base:= Min(p_salary, l_plafond_secu* 4)* 0.9825;
            if p_salary> l_plafond_secu* 4
              then l_base:= l_base+ (p_salary- l_plafond_secu* 4);

            l_base:= l_base+ l_prevoyance_pour_csg;

            l_employee_amount:= p_salary* m_employee_rate/ 100;
            _display_and_add(m_name, m_policy, l_base,
                m_employee_rate, l_employee_amount, 0, 0);

            if p_is_deductible
              then m_csg_deductible_sas:= l_employee_amount;
          end;
        end; // _compute_csg

      var l_c_save_display_strings: tStrings;
          l_sas_parameter_index: Integer;

      begin // compute_charges_sociales_sas
        l_c_save_display_strings:= s_c_strings;
        initialize_display(m_c_detail_result_list);

        l_total_employee:= 0; l_total_company:= 0;
        l_prevoyance_pour_csg:= 0;

        display(f_display('Salaire :', p_salary));
        display_line;

        with m_c_sas_parameter_list_2_ref do
        begin
          display_sas_parameter_list;
          // Exit;
          
          l_plafond_secu:= f_c_find_sas_parameter_by_name('plafond_secu_sas').m_employee_rate;

          for l_sas_parameter_index:= 0 to f_sas_parameter_count- 1 do
            with f_c_sas_parameter(l_sas_parameter_index) do
              if m_policy= 'T'
                then _compute_on_full_salary(f_c_self) else
              if m_policy= 'P'
                then _compute_on_plafond_secu(f_c_self) else
              if m_policy= 'B_1_8'
                then _compute_retraite_1_a_8_plafond(f_c_self) else
              if m_policy= 'T_<_4'
                then _compute_on_total_up_to_4_plafond(f_c_self) else
              if m_policy= 'T_1_8'
                then _compute_on_total_1_to_8_plafond(f_c_self) else
              if m_policy= 'T_<_8'
                then _compute_on_total_up_to_8_plafond(f_c_self) else
              if m_policy= 'CSG_d'
                then _compute_csg(f_c_self, k_deductible) else
              if m_policy= 'CSG'
                then _compute_csg(f_c_self, k_not_deductible)
                else
                  if m_policy= ''
                    then
                    else _display_and_add(m_name, m_policy, p_salary, 0, 0, 0, 0);
        end; // with m_c_sas_parameter_list_2_ref

        display_line;
        if p_salary= 0
          then p_salary:= 1;
        display(Format('total %s     %6.2f  %10s %s   %6.2f  %10s',
            [f_spaces(10+ 26+ 1), l_total_employee/ p_salary*100, f_10_2(l_total_employee),
             f_spaces(1), l_total_company/ p_salary* 100, f_10_2(l_total_company)]));

        display_line;
        m_total_charges_salaire_sas:= l_total_employee+ l_total_company;
        m_pourcent_charges_salariales:= (l_total_employee+ l_total_company)/ p_salary;
        display(Format('// overall %s      salaire %10s  charges %10s    %6.2f %%',
            [f_spaces(3), f_10_2(p_salary),
             f_10_2(l_total_employee+ l_total_company),
             m_pourcent_charges_salariales* 100]));

        display_line;
        display(f_display('// sal+ charges', p_salary+ m_total_charges_salaire_sas));
        display(f_display('// csg_deductible', m_csg_deductible_sas));

        initialize_display(l_c_save_display_strings);
      end; // compute_charges_sociales_sas

    Destructor c_calcul_charges_sas.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_calcul_charges_sas
    end. // u_c_calcul_charges_sas


