// 002 u_c_calcul_ir
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_ir;
  interface
    uses Classes, u_c_basic_object
        , u_strings, u_stringlist, u_c_stringlist

        , u_c_parameters

        ;

    type c_calcul_ir=
             class(c_basic_object)
               m_c_parameters_ref: c_parameters;
               m_c_detail_result_list: tStrings;
               m_title: String;

               m_month_count: Integer;

               m_tranche_count: Integer;
               m_seuil_array, m_taux_array: array of double;

               m_c_key_list, m_c_value_list: tStringList;

               m_retraites: Double;
               m_revenu_imposable: Double; // entered from form or from simulateur revenu
               m_nombre_parts: Double;
               m_csg_deductible_ir: Double;

               // -- SARL only
               m_dividendes_soumis_aux_charges_sociales,
                   m_dividendes_non_soumis_aux_charges_sociales: Double;

               m_IR, m_ir_precedent: Double;
               m_total_csg_crds, m_total_csg_crds_precedent: Double;

               // -- check our feuille d'imposition
               m_revenu_au_taux_forfaitaire: Double;
               m_delta_net: Double;

               m_total_charges: Double;

               Constructor create_calcul_ir(p_name: String; p_c_parameters_ref: c_parameters);

               function f_rate(p_key: String): Double;
               function f_amount(p_key: String): Double;

               procedure load_parameters(p_path, p_file_name: String);
               procedure display_parameters;

               procedure compute_ir;

               Destructor Destroy; Override;
             end; // c_calcul_ir

  implementation
    uses SysUtils, Math, ClipBrd
        , u_c_display, u_utils
        ;

    // -- c_calcul_ir

    const k_max= 10e15;

    function f_is_tranche(p_key: String): Boolean;
      begin
        Result:= Copy(p_key, 1, 8)= 'tranche_';
      end; // f_is_tranche

    Constructor c_calcul_ir.create_calcul_ir(p_name: String; p_c_parameters_ref: c_parameters);
      begin
        Inherited create_basic_object(p_name);
        m_c_parameters_ref:= p_c_parameters_ref;

        m_title:= 'IMPOT SUR LE REVENU';

        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;

        m_nombre_parts:= 1;
      end; // create_calcul_ir

    procedure c_calcul_ir.load_parameters(p_path, p_file_name: String);
      var l_c_stinglist: c_stringlist;

      procedure _add_key_value(p_key, p_value: String);
        begin
          display('add '+ p_key+ ' '+ p_value+ '<');
          m_c_key_list.Add(p_key);
          m_c_value_list.Add(p_value);
        end; // _add_key_value

      var l_key, l_seuil, l_taux, l_value: String;

      begin // load_parameters
        l_c_stinglist:= c_stringlist.create_stringlist('', p_path, p_file_name);

        with l_c_stinglist do
        begin
          if m_c_stringlist.Count= 0
            then begin
                m_c_parameters_ref.add_error(p_file_name+ ' IR_parameter_file_empty');
                Free;
                Exit;
             end;
          stop_if_empty;

          // -- a first bogus tranche "0 0"
          m_tranche_count:= 1;
          SetLength(m_seuil_array, m_tranche_count);
          SetLength(m_taux_array, m_tranche_count);

          while not f_end_of_file do
          begin
            if not f_is_blank_or_comment
              then begin
                  // NO _add_key_value(f_next_value, f_next_value);
                  l_key:= f_next_value;

                  if f_is_tranche(l_key)
                    then begin
                        l_seuil:= f_next_value; l_taux:= f_next_value;
                        Inc(m_tranche_count);
                        SetLength(m_seuil_array, m_tranche_count);
                        SetLength(m_taux_array, m_tranche_count);
                        if l_seuil= '+'
                          then m_seuil_array[m_tranche_count- 1]:= k_max
                          else m_seuil_array[m_tranche_count- 1]:= StrToFloat(l_seuil);
                        m_taux_array[m_tranche_count- 1]:= StrToFloat(l_taux);

                        l_value:= l_seuil+ ' '+ l_taux;
                      end
                    else begin
                        // -- taux forfaitaire
                        l_value:= f_next_value;
                      end  ;

                  _add_key_value(l_key, l_value);
                end
              else
                if m_title= ''
                  then m_title:= m_trimmed_line
                  else _add_key_value('','');

            read_line;
          end; // while not f_end_of_file
        end; // with l_c_stringlist

        display(m_c_key_list.Text);
        display(m_c_value_list.Text);
      end; // load_parameters

    function c_calcul_ir.f_rate(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_rate

    function c_calcul_ir.f_amount(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of])
      end; // f_amount

    procedure c_calcul_ir.display_parameters;
      var l_ir_parameter_index, l_tranche_index: Integer;
          l_key, l_seuil: String;
          // l_key_value_index: Integer;
      begin
        display_line;
        display(m_title);

        l_tranche_index:= 1;

        with m_c_key_list do
          for l_ir_parameter_index:= 0 to Count- 1 do
          begin
            l_key:= Strings[l_ir_parameter_index];
            if l_key= ''
              then begin
                  display_line;
                end
              else
                if f_is_tranche(l_key)
                  then begin
                      if m_seuil_array[l_tranche_index]= k_max
                        then l_seuil:= f_spaces(10)
                        else l_seuil:= Format('%10.2f', [m_seuil_array[l_tranche_index] ]);
                      display(Format('%-10s   %8s  %5.2f',
                          [l_key, l_seuil, m_taux_array[l_tranche_index] ]));
                      Inc(l_tranche_index);
                    end
                  else begin
                       display(Format('%-27s   %10.2f %5.2f',
                            [l_key, f_amount(l_key), f_rate(l_key) ]))  ;
                    end;
          end; // fpr
      end; // display_parameters

    procedure c_calcul_ir.compute_ir;
      var l_micro_foncier_net: Double;

      procedure _calcul_csg_crds_loyers;
        var l_base_flat_tax: Double;

        procedure _calcule_dividendes;
          begin
            with m_c_parameters_ref do
            begin
              m_dividendes_associe:= 0;
              m_dividendes_non_soumis_aux_charges_sociales:= 0;
              m_dividendes_soumis_aux_charges_sociales:= 0;

              if m_dividendes> 0
                then begin
                    display('---------- >');
                    display(f_display('dividendes', m_dividendes));
                    m_dividendes_associe:= m_dividendes* m_pourcent_capital;
                    display(f_display('dividende_associ', m_dividendes_associe)
                      + Format('%5s %5.2f * %s',
                          ['', m_pourcent_capital, f_10_2(m_dividendes)]));
                    if m_status_kind= e_sarl_status
                      then begin
                          m_dividendes_non_soumis_aux_charges_sociales:= Min(m_dividendes_associe,
                              f_rate('seuil_dividendes_sarl_ir')/ 100* m_pourcent_capital* m_capital);
                          m_dividendes_soumis_aux_charges_sociales:= m_dividendes_associe- m_dividendes_non_soumis_aux_charges_sociales;

                          display_line;
                          display(f_display('divid_soumis_charges',     m_dividendes_soumis_aux_charges_sociales));
                          display(f_display('divid_non_soumis_charges', m_dividendes_non_soumis_aux_charges_sociales));
                        end;
                        
                    display('---------- <');
                    display_line;
                 end
            end; // with m_c_parameters_ref
          end; // _calcule_dividendes_sarl

        procedure __display_and_add_csg(p_title: String; p_rate, p_base: Double);
          var l_amount: Double; // l_rate,
          begin
            if p_base> 0
              then begin
                  l_amount:= p_base* p_rate/ 100;
                  display(Format('%-35s %10s  %5.2f %%  %10s',
                      [p_title, f_10_2(p_base), p_rate, f_10_2(l_amount)]));
                  m_total_csg_crds:= m_total_csg_crds+ l_amount;
                end;
          end; // __display_and_add_csg

        var l_flat_tax_csg: Double;
            l_full_flat_tax: Double;
            l_csg_crds_sas_IR: Double;
            l_dividendes_sarl_ou_sas: Double;

        begin // _calcul_csg_crds_loyers
          with m_c_parameters_ref do
          begin
            _calcule_dividendes;

            m_total_csg_crds:= 0;

            display(f_display('micro_foncier_net', l_micro_foncier_net));

            l_flat_tax_csg:= f_rate('flat_tax_CSG')+ f_rate('flat_tax_CRDS')+ f_rate('flat_tax_solidarite');

            display_line;
            __display_and_add_csg('flat_tax_csg_loyer', l_flat_tax_csg, l_micro_foncier_net);
            if m_status_kind= e_sarl_status
              then begin
                  __display_and_add_csg('csg_crds_divid_assoc_non_soumis', l_flat_tax_csg, m_dividendes_non_soumis_aux_charges_sociales);
                end;

            display_line;

            l_csg_crds_sas_IR:= f_rate('flat_tax_CSG')+ f_rate('flat_tax_CSG')+ f_rate('flat_tax_CRDS');
            __display_and_add_csg('csg_crds_SAS_IR', l_csg_crds_sas_IR, m_revenu_sas_IR);

            l_base_flat_tax:= 0;
            display_line;
            display(f_display('interets_associes', m_interets_compte_associe));

            if m_status_kind= e_sarl_status
              then begin
                  display(f_display('csg_crds_divid_assoc_soum', m_dividendes_soumis_aux_charges_sociales));
                  l_dividendes_sarl_ou_sas:= m_dividendes_soumis_aux_charges_sociales;
                end
              else begin
                  display(f_display('dividendes_associ', m_dividendes_associe));
                  l_dividendes_sarl_ou_sas:= m_dividendes_associe;
                end;

            l_base_flat_tax:= m_interets_compte_associe+  l_dividendes_sarl_ou_sas;

            display_line;
            display(f_display(' = base_csg_crds', l_base_flat_tax));
            display_line;

            l_full_flat_tax:= l_flat_tax_csg+ f_rate('flat_tax_taux_forfaitaire');
            __display_and_add_csg('flat_tax', l_full_flat_tax, l_base_flat_tax);

            display_line;
            display(f_display_2('total_csg_flat_tax', 29, m_total_csg_crds));
          end; // with m_c_parameters_ref
        end; // _calcul_csg_crds_loyers

      var l_total_revenus: Double;

      procedure _display_and_add_revenu(p_title: String; p_amount: Double);
        var l_pourcent: String;
        begin
          if p_amount> 0
            then begin
                if p_title= 'salaire'
                  then l_pourcent:= Format('%5s %s', ['', f_10_2(p_amount/ (p_amount+ + m_c_parameters_ref.m_dividendes_associe))])
                  else l_pourcent:= '';
                display(f_display(p_title, p_amount)+ l_pourcent);

                l_total_revenus:= l_total_revenus+ p_amount;
              end;
        end; // _display_and_add_revenu

      var l_total_IR_CSG: Double;

      procedure _display_and_add_impots(p_title: String; p_amount: Double);
        var l_display: String;
        begin
          l_display:= f_display(p_title, p_amount);

          if m_ir_precedent<> 0
            then
              if pos('IR', p_title)> 0
                then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), p_amount- m_ir_precedent])
                else
                  if Pos('CSG', p_title)> 0
                    then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), p_amount- m_total_csg_crds_precedent]);

          display(l_display);

          l_total_IR_CSG:= l_total_IR_CSG+ p_amount;
        end; // _display_and_add_impots

      var l_charges_IS_IR: Double;

      procedure _display_and_add(p_title: String; p_charge: Double);
        begin
          display(f_display(p_title, p_charge));
          l_charges_IS_IR:= l_charges_IS_IR+ p_charge;
        end; // _display_and_add

      procedure _generate_summary;
        var l_c_summary_list: tStringList;
            l_reserves_associe_futures: Double;
            l_reste_net: Double;
            l_somme_revenus: Double;
        begin
          l_reserves_associe_futures:= 0;
          l_c_summary_list:= Nil;

          with m_c_parameters_ref do
          begin
            if m_insert_result_index> 0
              then begin
                  l_c_summary_list:= tStringList.Create;
                  initialize_display(l_c_summary_list);
                end;

            // add_table_start('3_ir', m_title_net);
            add_table_start_2('3_ir', m_title_net, '15', '#CCFFFF');

            if (m_reserve> 0) and (m_status_kind<> e_sas_IR_status)
              then begin
                  display('---------- >');
                  display(f_display('rserves', m_reserve));
                  l_reserves_associe_futures:= m_reserve* m_pourcent_capital;
                  display(f_display('rserve_associ', l_reserves_associe_futures)
                     + Format('%s %s * %5.2f', ['', f_10_2(m_reserve), m_pourcent_capital]));

                  display(f_display('rserve_associ_future', l_reserves_associe_futures* 0.70)
                     + Format('%s %s * %5.2f', ['', f_10_2(l_reserves_associe_futures), 0.70]));
                  display('---------- <');
                  l_reserves_associe_futures:= l_reserves_associe_futures* 0.70;
                  display_line;
                end;

            display('--- REVENUS foyer fiscal');
            l_total_revenus:= 0;
            _display_and_add_revenu('salaire', m_salaire);
            _display_and_add_revenu(' + retraites', m_retraites);
            _display_and_add_revenu(' + micro_foncier_brut', m_micro_foncier_brut);
            _display_and_add_revenu(' + intrts_compte_assoc', m_interets_compte_associe);
            _display_and_add_revenu(' + dividendes_associ', m_dividendes_associe);

            if m_status_kind<> e_sas_IR_status
              then _display_and_add_revenu(' + rserves_assoc_futures', l_reserves_associe_futures);

            display_line;
            display(f_display('= total_revenus', l_total_revenus));

            display_line;
            display('--- PRELEVEMENTS foyer fiscal');
            l_total_IR_CSG:= 0;
            _display_and_add_impots('IR', m_IR);
            _display_and_add_impots(' + CSG', m_total_csg_crds);

            display_line;
            display(f_display('= total_prlvements', l_total_IR_CSG));

            display_line;
            l_reste_net:= l_total_revenus- l_total_IR_CSG;
            display(f_display('RESTE_NET', l_reste_net)
                + Format('   prel / reste_net    %3.2f %% ', [l_total_IR_CSG/ l_reste_net* 100]));
            ClipBoard.AsText:= f_10_2(l_reste_net);

            // -- now insert at the start
            if m_insert_result_index> 0
              then begin
                  m_c_result_list_ref.Insert(m_insert_result_index, l_c_summary_list.Text);
                  l_c_summary_list.Free;
                  initialize_display(m_c_result_list_ref);
                  display_debug('xxx_reseet_main_result');
                end;

          end; // with m_c_parameters_ref
        end; // _generate_summary

      var l_c_save_display_strings: tStrings;

          l_revenu_net_imposable: Double;
          l_revenu_brut_global: Double;
          l_revenu_imposable: Double;

          l_revenu_imposable_par_part: double;
          l_impot_par_tranches: Double;

          l_impot_taux_forfaitaire: Double;

          l_tranche_index: Integer;
          l_base, l_amount, l_total: Double;
          l_seuil: String;

      begin // compute_ir
        with m_c_parameters_ref do
        begin
          l_impot_taux_forfaitaire:= 0;

          add_table_start('5_ir', m_title);

          display(f_display('salaires', m_revenu_imposable));
          if m_csg_deductible_ir> 0
            then display(f_display(' + csg_deductible', m_csg_deductible_ir));
          display(f_display(' + retraites', m_retraites));
          m_revenu_imposable:= m_revenu_imposable+ m_csg_deductible_ir+ m_retraites;

          display_line;

          l_revenu_net_imposable:= 0.90* m_revenu_imposable;
          l_micro_foncier_net:= m_micro_foncier_brut* 0.70;
          l_revenu_brut_global:= l_revenu_net_imposable+ l_micro_foncier_net;
          l_revenu_imposable:= l_revenu_brut_global- m_csg_deductible_ir;

          display(f_display('=  revenu_imposable', m_revenu_imposable));
          display(f_display(' - frais_10_%', 0.10* m_revenu_imposable));
          display(f_display('=  revenu_net_imposable', l_revenu_net_imposable));
          display(f_display(' + micro_foncier_net', l_micro_foncier_net)
               + Format(' (0.70 * foncier %10f.3)', [m_micro_foncier_brut]));
          display(f_display('= revenu_brut_global', l_revenu_brut_global));
          display(f_display(' - csg_deductible', m_csg_deductible_ir));
          display(f_display('= revenu_imposable', l_revenu_imposable));
          if m_delta_net> 0
            then display(f_display('delta_net_??? ', m_delta_net));

          display_line;

          l_total:= 0;
          l_revenu_imposable:= l_revenu_imposable+ m_delta_net;

          l_c_save_display_strings:= s_c_strings;
          initialize_display(m_c_detail_result_list);

          display(f_display('revenu_imposable', l_revenu_imposable));
          display_line;

          l_revenu_imposable_par_part:= l_revenu_imposable/ m_nombre_parts;
          display(f_display('revenu_imposable_par_part', l_revenu_imposable_par_part));
          display_line;
          // l_revenu_imposable_par_part:= 31000* 0.9;

          l_impot_par_tranches:= 0;

          for l_tranche_index:= 1 to m_tranche_count- 1 do
            if l_revenu_imposable_par_part> m_seuil_array[l_tranche_index- 1]
              then begin
                  l_base:= Min(l_revenu_imposable_par_part, m_seuil_array[l_tranche_index]);
                  l_base:= l_base- m_seuil_array[l_tranche_index- 1];
                  l_amount:= l_base* m_taux_array[l_tranche_index]/ 100;
                  if m_seuil_array[l_tranche_index]= k_max
                    then l_seuil:= f_spaces(10)
                    else l_seuil:= Format('%10s', [f_10_2(m_seuil_array[l_tranche_index]) ]);
                  display(Format('%d %10s  %10s  %6.2f    %10s',
                      [l_tranche_index, l_seuil, f_10_2(l_base), m_taux_array[l_tranche_index], f_10_2(l_amount)]));

                  l_total:= l_total+ l_amount;
                end;

          display_line;
          display(Format('%-35s %10s', ['total', f_10_2(l_total)]));

          m_IR:= l_total* m_nombre_parts;
          display_line;
          // display(f_display('rev_imp_xxx', l_revenu_imposable));
          display(Format('total_pour_%1.1f_parts %s %10s   %5.2f %%',
              [m_nombre_parts, f_spaces(14), f_10_2(m_IR), m_IR/ l_revenu_imposable* 100] ));

          initialize_display(l_c_save_display_strings);
          display_line;
          display(f_display('impot_a_payer', m_IR));

          l_total:= l_impot_par_tranches+ l_impot_taux_forfaitaire;

          display_line;

          initialize_display(m_c_detail_result_list);

          _calcul_csg_crds_loyers;

          initialize_display(l_c_save_display_strings);

          _generate_summary;
        end; // with m_c_parameters_ref
      end; // compute_ir

    Destructor c_calcul_ir.Destroy;
      begin
        m_c_key_list.Free;
        m_c_value_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_calcul_ir
    end. // u_c_calcul_ir


