// 002 u_c_calcul_is
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_calcul_is;
  interface
    uses Classes, u_c_basic_object
        , u_c_parameters
        ;

    const k_set_dividendes_to_benefice_net_true= True;
          k_set_dividendes_to_benefice_net_false= False;

    type c_calcul_is=
             class(c_basic_object)
               m_c_parameters_ref: c_parameters;
               m_c_detail_result_list: tStrings;
               m_title: String;

               m_tranche_is_count: integer;
               m_seuil_is_array, m_taux_is_array: array of Double;

               m_c_key_list, m_c_value_list: tStringList;

               m_interets_associes_deductibles, m_interets_associes_non_deductibles: Double;

               m_resultat_moins_salaires: Double;

               m_IS, m_taux_is_moyen: Double;

               m_benefice_avant_is, m_benefice_net: Double;

               m_benefice_net_precedent: Double;

               Constructor create_calcul_is(p_name: String; p_c_parameters_ref: c_parameters);

               procedure load_is_parameters_from_file(p_path, p_file_name: String);
               function f_amount(p_key: String): Double;
               function f_rate(p_key: String): Double;
               procedure display_is_parameters;

               procedure calcule_is_interets_deductibles;
               procedure calcule_is_sur_revenu_imposable(p_set_dividendes_to_benefice_net: Boolean);

               Destructor Destroy; Override;
             end; // c_calcul_is

  implementation
    uses SysUtils, Math, TypInfo
        , u_c_display
        , u_c_stringlist

        , u_utils
        ;

    const k_max= 10e15;

    // -- c_calcul_is

    Constructor c_calcul_is.create_calcul_is(p_name: String; p_c_parameters_ref: c_parameters);
      begin
        Inherited create_basic_object(p_name);
        m_c_parameters_ref:= p_c_parameters_ref;

        m_title:= 'IMPOTS SUR LES SOCIETES';
        m_c_key_list:= tStringList.Create;
        m_c_value_list:= tStringList.Create;
      end; // create_calcul_is

    procedure c_calcul_is.load_is_parameters_from_file(p_path, p_file_name: String);

      procedure _add_key_value(p_key, p_value: String);
        begin
          display('add '+ p_key+ ' '+ p_value);
          m_c_key_list.Add(p_key);
          m_c_value_list.Add(p_value);
        end; // _add_key_value

      var l_seuil, l_taux: String;
          l_key, l_value: String;

      begin // load_is_parameters_from_file
        // -- a first bogus (0, 0) slice
        m_tranche_is_count:= 1;
        SetLength(m_seuil_is_array, m_tranche_is_count);
        SetLength(m_taux_is_array, m_tranche_is_count);

        with c_stringlist.create_stringlist('', p_path, p_file_name) do
        begin
          if m_c_stringlist.Count= 0
            then begin
                m_c_parameters_ref.add_error(p_file_name+ ' IS_parameter_file_empty');
                Free;
                Exit;
             end;

          while not f_end_of_file do
          begin
            display(m_the_line);
            if not f_is_blank_or_comment
              then begin
                  if Pos('tranche_is_', m_the_line)> 0
                    then begin
                        f_next_value;
                        l_seuil:= f_next_value; l_taux:= f_next_value;
                        Inc(m_tranche_is_count);
                        SetLength(m_seuil_is_array, m_tranche_is_count);
                        SetLength(m_taux_is_array, m_tranche_is_count);
                        if l_seuil= '+'
                          then m_seuil_is_array[m_tranche_is_count- 1]:= k_max
                          else m_seuil_is_array[m_tranche_is_count- 1]:= StrToFloat(l_seuil);
                        m_taux_is_array[m_tranche_is_count- 1]:= StrToFloat(l_taux);
                      end
                    else begin
                        l_key:= f_next_value; l_value:= f_next_value;
                        _add_key_value(l_key, l_value);
                      end;
                end
              else
                if m_title= ''
                  then m_title:= m_trimmed_line
                  else _add_key_value('','');
            read_line;
          end; // while not f_end_of_file

          Free;
        end; // with c_stringlist
      end; // load_is_parameters_from_file

    function c_calcul_is.f_amount(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              // display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_amount

    function c_calcul_is.f_rate(p_key: String): Double;
      var l_index_of: integer;
      begin
        l_index_of:= m_c_key_list.IndexOf(p_key);
        if l_index_of< 0
          then begin
              // display_bug_stop('no_key '+ p_key);
              m_c_parameters_ref.add_error('no_key '+ p_key);
              Result:= 0;
            end
          else Result:= StrToFloat(m_c_value_list[l_index_of]);
      end; // f_rate

    procedure c_calcul_is.display_is_parameters;
      var l_tranche_is: Integer;
          l_is_parameter_index: Integer;
          l_key: String;
      begin
        display(m_title);

        for l_tranche_is:= 0 to m_tranche_is_count- 1 do
          display(Format('tranche_is_%d %10.2f %5.2f',
            [l_tranche_is, m_seuil_is_array[l_tranche_is], m_taux_is_array[l_tranche_is] ]));
        display_line;

        with m_c_key_list do
          for l_is_parameter_index:= 0 to Count- 1 do
          begin
            l_key:= Strings[l_is_parameter_index];
            if l_key= ''
              then display_line
              else display(Format('%-25s %10.2f', [l_key, f_amount(l_key)]));
          end;
      end; // display_is_parameters

    procedure c_calcul_is.calcule_is_interets_deductibles;
      var l_taux_deductible: Double;
      begin
        with m_c_parameters_ref do
        begin
          if m_benefice_avant_is> 0
            then begin
                m_interets_compte_associe:= m_taux_interets_compte_associe* m_capital;

                l_taux_deductible:= Min(m_taux_interets_compte_associe, f_rate('taux_interet_moyen_pret')/ 100);
                m_interets_associes_deductibles:= l_taux_deductible* m_capital;
                m_interets_associes_non_deductibles:= Max(0, m_interets_compte_associe- m_interets_associes_deductibles);

                if m_interets_associes_non_deductibles> 0
                  then begin
                      display('---------- >');
                      display(f_display(' + intrts_associ_dduct'  , m_interets_associes_deductibles));
                      display(f_display(' + intrts_non_dduct'      , m_interets_associes_non_deductibles));
                      display(f_display(' = intrtst'                , m_interets_compte_associe));
                      display('---------- <');
                      display_line;
                    end;
              end;
        end; // with m_c_parameters_ref
      end; // calcule_is_interets_deductibles

    procedure c_calcul_is.calcule_is_sur_revenu_imposable(p_set_dividendes_to_benefice_net: Boolean);
      var l_c_save_display_strings: tStrings;

          l_tranche_is: Integer;
          l_base, l_taux_tranche: Double;
          l_is_tranche: Double;
          l_display: String;
      begin
        l_c_save_display_strings:= s_c_strings;
        initialize_display(m_c_detail_result_list);
        clear_display;

        with m_c_parameters_ref do
        begin
          if m_benefice_avant_is> 0
            then begin
                display(f_display('rsu_moins_salaires'     , m_benefice_avant_is));
                display_line;

                m_benefice_avant_is:= m_benefice_avant_is- m_micro_foncier_brut- m_interets_associes_deductibles;

                display(f_display(' - micro_foncier'           , m_micro_foncier_brut));
                display(f_display(' - intrts_associ_dduct' , m_interets_associes_deductibles));
                display(f_display(' = bnf_soumis__is'       , m_benefice_avant_is));

                m_taux_is_moyen:= 0;

                if m_status_kind= e_sas_ir_status
                  then m_IS:= 0
                  else
                    begin
                      display_line;

                      m_IS:= 0;
                      for l_tranche_is:= 1 to m_tranche_is_count- 1 do
                      begin
                        l_base:= m_benefice_avant_is- m_seuil_is_array[l_tranche_is- 1];
                        if l_base> 0
                          then begin
                              l_base:= Min(m_benefice_avant_is, m_seuil_is_array[l_tranche_is]);
                              l_base:= l_base- m_seuil_is_array[l_tranche_is- 1];

                              l_taux_tranche:= m_taux_is_array[l_tranche_is] ;
                              l_is_tranche:= l_base* l_taux_tranche / 100;
                              display(Format('is_1 %10s %10s  %5.2f   %10s',
                                  [f_10_2(m_seuil_is_array[l_tranche_is]), f_10_2(l_base), l_taux_tranche, f_10_2(l_is_tranche)]));

                              m_IS:= m_IS+ l_is_tranche;
                            end;
                      end; // for l_tranche_is

                    end;
                display_line;
                display(Format('total_is %s %10s  %5.2f %%  ',
                    [f_spaces(10+ 16), f_10_2(m_IS), m_IS/ m_benefice_avant_is]));
                m_benefice_net:= m_benefice_avant_is- m_IS;

                display_line;

                display(f_display('imposable__IS'             , m_benefice_avant_is));
                display(f_display(' - IS', m_IS));
                display(f_display(' - intrts_non_dduct', m_interets_associes_non_deductibles));
                m_benefice_net:= m_benefice_net- m_interets_associes_non_deductibles;

                // -- calcul prlevement tt
                m_IS_:= m_IS;

                l_display:= f_display(' = bnfices_net',   m_benefice_net);
                if m_benefice_net_precedent<> 0
                  then l_display:= l_display+ Format('%s [%10.2f]', [f_spaces(2), m_benefice_net- m_benefice_net_precedent]);
                display(l_display);

                display_line;

                if p_set_dividendes_to_benefice_net
                  then begin
                      m_dividendes:= m_benefice_net;
                      m_reserve:= 0;
                    end
                  else m_reserve:= m_benefice_net- m_dividendes;

                display(f_display(' - dividendes', m_dividendes));

                    begin
                      display_line;
                      display(f_display(' = rserves', m_reserve));
                    end;

                m_taux_is_moyen:= m_IS / m_benefice_avant_is;
                initialize_display(l_c_save_display_strings);
              end
            else begin
                l_c_save_display_strings:= s_c_strings;
                initialize_display(m_c_detail_result_list);
                clear_display;
                display(f_display('*** revenu_avant_is< 0 %10.2', m_benefice_avant_is));
                initialize_display(l_c_save_display_strings);
              end;
        end; // with m_c_parameters_ref
      end; // calcule_is_sur_revenu_imposable

    Destructor c_calcul_is.Destroy;
      begin
        m_c_key_list.Free;
        m_c_value_list.Free;

        Inherited;
      end; // Destroy

    begin // u_c_calcul_is
    end. // u_c_calcul_is


