// 002 u_c_parameters
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_parameters;
  interface
    uses Classes, u_c_basic_object
        ;

    type t_status_kind= (e_unknown_status, e_sarl_status, e_sas_status, e_sas_ir_status);

         c_parameters=
             class(c_basic_object)
               m_c_result_list_ref: tStrings;

               m_status_kind: t_status_kind;

               // -- IS
               m_capital, m_pourcent_capital: Double;
               m_dividendes, m_dividendes_associe: Double;

               m_revenu_sas_IR: Double;
               m_taux_interets_compte_associe, m_interets_compte_associe: Double;

               m_micro_foncier_brut: Double;
               m_reserve: Double;

               m_salaire: Double;
               m_resultat_exploitation: Double;

               m_title_net: String;
               m_insert_result_index: Integer;
               m_IS_: Double;
               m_debug_display: Boolean;

               m_error_string: String;

               Constructor create_parameters(p_name: String);

               procedure add_table_start(p_marker, p_title: String);
               procedure add_table_start_2(p_marker, p_title, p_padding, p_color: String);

               procedure add_line(p_text: String);

               procedure display_debug(p_text: String);

               procedure reset_errors;
               procedure add_error(p_error: String);
               function f_has_error: Boolean;

               Destructor Destroy; Override;
             end; // c_parameters

    function f_status_name(p_status_type: t_status_kind): String;
    function f_type_societe(p_societe: String): t_status_kind;

  implementation
    uses SysUtils, TypInfo
        , u_c_display
        , u_utils
        ;

    function f_status_name(p_status_type: t_status_kind): String;
      begin
        Result:= GetEnumName(TypeInfo(t_status_kind), Integer(p_status_type));
        Delete(Result, 1, 2);
        Delete(Result, Length(Result)- Length('_status')+ 1, Length('_status'));
      end; // f_status_name

    function f_type_societe(p_societe: String): t_status_kind;
      begin
        if p_societe= 'sarl'
          then Result:= e_sarl_status else
        if p_societe= 'sas'
          then Result:= e_sas_status else
        if p_societe= 'sas_ir'
          then Result:= e_sas_ir_status else
        Result:= e_unknown_status;
      end; // f_type_societe

    // -- c_parameters

    Constructor c_parameters.create_parameters(p_name: String);
      begin
        Inherited create_basic_object(p_name);
        m_title_net:= 'RESULTAT NET';
      end; // create_parameters

    procedure c_parameters.add_table_start_2(p_marker, p_title, p_padding, p_color: String);
      begin
        with m_c_result_list_ref do
        begin
          if p_title<> ''
            then begin
                display(p_title);
                display_line;
              end;
        end; // with m_c_result_list_ref
      end; // add_table_start_2

    procedure c_parameters.add_table_start(p_marker, p_title: String);
      begin
        with m_c_result_list_ref do
        begin
          if p_title<> ''
            then begin
                display(p_title);
                display_line;
              end;
        end; // with m_c_result_list_ref
      end; // add_table_start

    procedure c_parameters.add_line(p_text: String);
      begin
        with m_c_result_list_ref do
          display('    '+ p_text+ '<BR>');
      end; // add_line

    procedure c_parameters.display_debug(p_text: String);
      begin
        if m_debug_display
          then display(p_text);
      end; // display_debug

    procedure c_parameters.reset_errors;
      begin
        m_error_string:= '';
      end; // reset_error

    procedure c_parameters.add_error(p_error: String);
      begin
        m_error_string:= m_error_string+ '<BR>'+ p_error;
      end; // add_error

    function c_parameters.f_has_error: Boolean;
      begin
        Result:= m_error_string<> '';
      end; // f_has_error

    Destructor c_parameters.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_parameters
    end. // u_c_parameters


