// 002 u_c_simulateur_revenus
// 28 feb 2020

// -- (C) John COLIBRI 2020
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_simulateur_revenus;
  interface
    uses Classes, u_c_basic_object

        , u_c_parameters

        , u_c_calcul_charges_sarl

        , u_c_sas_parameter_list_2
        , u_c_calcul_charges_sas

        , u_c_calcul_is

        , u_c_calcul_ir
        ;

    type
         c_simulateur_revenus=
             class(c_basic_object)
               m_c_parameters_ref: c_parameters;
               m_title: String;

               m_c_calcul_charges_sarl_ref: c_calcul_charges_sarl;
               m_c_calcul_charges_sas_ref: c_calcul_charges_sas;
               m_c_calcul_is_ref: c_calcul_is;
               m_c_calcul_ir_ref: c_calcul_ir;

               m_cotisations_sociales_obligatoire_precedentes: Double;

               m_objectif_pourcent_salaire, m_objectif_renvenu_net_societe: Double;

               Constructor create_simulateur_revenus(p_name: String; p_c_parameters_ref: c_parameters);

               procedure compute_revenu_sarl(p_c_result_strings: tStrings);
               procedure compute_revenu_sas(p_c_result_strings: tStrings);
               procedure compute_revenu_sas_IR(p_c_result_strings: tStrings);

               Destructor Destroy; Override;
             end; // c_simulateur_revenus

  implementation
    uses SysUtils, u_c_display
        , u_utils
        ;

    // -- c_simulateur_revenus

    Constructor c_simulateur_revenus.create_simulateur_revenus(p_name: String; p_c_parameters_ref: c_parameters);
      begin
        Inherited create_basic_object(p_name);
        m_c_parameters_ref:= p_c_parameters_ref;

        m_title:= 'simulateur_revenus';

        m_objectif_pourcent_salaire:= -1;
      end; // create_simulateur_revenus

    procedure c_simulateur_revenus.compute_revenu_sarl(p_c_result_strings: tStrings);

      procedure _calcul_litteral;
        begin
          with m_c_parameters_ref do
          begin
            with m_c_calcul_charges_sarl_ref do
            begin
              add_table_start('100_rev', m_title);

              compute_charges_sociales_sarl(m_salaire, m_cotisations_sociales_obligatoire_precedentes);
            end; // with m_c_calcul_charges_sarl_ref

            with m_c_calcul_is_ref do
            begin
              add_table_start('200_ref', m_title);

              m_benefice_avant_is:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl;

              calcule_is_interets_deductibles;

              display(f_display('revenu_brut', m_resultat_exploitation));
              display(f_display(' - salaires', m_salaire));
              display(f_display(' - charges salariales', m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl));
              display(f_display(' = resu_moins_salaires', m_benefice_avant_is));
              display_line;

              m_status_kind:= e_sarl_status;

              calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_false);
            end; // with m_c_calcul_is_ref

            with m_c_calcul_ir_ref do
            begin
              m_total_charges:= m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl;

              m_revenu_imposable:= m_salaire;
              m_csg_deductible_ir:= m_c_calcul_charges_sarl_ref.m_csg_deductible_sarl;

              compute_ir;
            end;
          end; // with m_c_parameters_ref
        end; // _calcul_litteral

      procedure _calcul_iteratif;
        const k_display_false= False;
              k_display_true= True;
        var l_revenu_simple: Double;

        procedure _calcul_zero_salaire;
          // var l_taux_is: Double;
          begin
            with m_c_parameters_ref do
            begin
              // -- pas de salaire  SARL
              m_salaire:= 0;

              with m_c_calcul_is_ref do
              begin
                add_table_start('46_rev_IT', m_title);

                m_resultat_moins_salaires:= m_resultat_exploitation;
                m_benefice_avant_is:= m_resultat_moins_salaires;

                calcule_is_interets_deductibles;

                m_resultat_moins_salaires:= m_resultat_exploitation;

                calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_True);

                display(Format('obj_ %8.2f', [m_dividendes]));

              end; // with m_c_calcul_is_ref

              m_salaire:= 0;
            end; // with m_c_parameters_ref
          end; // _calcul_zero_salaire

        procedure _display(p_text: String);
          begin
            if k_display_true
              then display(p_text);
          end; // _display

        procedure _calcul_100_salaire;
          var l_pourcent_charges: Double;

          procedure _itere_calcul(p_iteration: Integer; p_display: Boolean);
            var l_delta: Double; // , l_taux_is
            begin
              with m_c_parameters_ref do
              begin
                if p_display= k_display_false
                  then push_stop_display(1);

                _display('');
                _display('---');
                with m_c_calcul_charges_sarl_ref do
                begin
                  add_table_start('40_rev_IT', m_title);

                  compute_charges_sociales_sarl(m_salaire, m_cotisations_sociales_obligatoire_precedentes);

                  l_pourcent_charges:= m_pourcent_charges_salariales;
                end;

                display_debug(Format('exp %8.2f i %5.2f loy %8.2f sal %8.2f chge %8.2f solde %8.2f',
                    [m_resultat_exploitation, m_interets_compte_associe, m_micro_foncier_brut,
                     m_salaire, m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl,
                     m_resultat_exploitation- m_interets_compte_associe- m_micro_foncier_brut
                     - m_salaire- m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl]));

                with m_c_calcul_is_ref do
                begin
                  add_table_start('46_rev_IT', m_title);

                  m_benefice_avant_is:= m_resultat_moins_salaires;

                  calcule_is_interets_deductibles;

                  m_resultat_moins_salaires:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl;
                  display(Format('ex %10.f  sal %10.2f chges %10.2f  sal+chge %10.2f',
                      [m_resultat_exploitation, m_salaire, m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl,
                       m_salaire+ m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl]));
                  // display(Format('resu_moins_salaires %s %10.2f', [f_spaces(7), m_resultat_moins_salaires]));

                  calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_true);

                  l_delta:= m_benefice_net/ (1- m_taux_is_moyen)/ (1 + m_c_calcul_charges_sarl_ref.m_pourcent_charges_salariales);

                  m_salaire:= m_salaire+ l_delta;

                  display_debug(Format('  iter %2d  ben_net %8.2f pourc_charges %5.2f taux_is %5.2f delta %5.2f sal %8.2f',
                      [p_iteration, m_benefice_net,
                       m_c_calcul_charges_sarl_ref.m_pourcent_charges_salariales, m_taux_is_moyen, m_salaire]));
                end; // with m_c_calcul_is_ref

                if p_display= k_display_false
                  then pop_stop_display;
              end; // with m_c_parameters_ref
            end; // _itere_calcul

          var l_iteration: Integer;

          begin // _calcul_100_salaire
            with m_c_parameters_ref do
            begin
              l_pourcent_charges:= 0.45;
              m_salaire:= l_revenu_simple/ (1+ l_pourcent_charges);
              l_iteration:= 0;

              repeat
                _itere_calcul(l_iteration, k_display_false);

                Inc(l_iteration);
              until l_iteration> 5;

              _itere_calcul(l_iteration, k_display_true);

              m_dividendes:= 0;
            end; // with m_c_parameters_ref
          end; // _calcul_100_salaire

        procedure _calcul_proportionnel;
          var l_pourcent_charges, l_taux_is: Double;

          function _f_salaire(p_pourcent_charges: Double): Double;
            var l_diviseur: Double;
            begin
              with m_c_parameters_ref do
              begin
                with m_c_calcul_is_ref do
                begin
                  l_diviseur:= (1- m_objectif_pourcent_salaire)/ m_objectif_pourcent_salaire
                      + (1+ l_pourcent_charges)* (1- l_taux_is);

                  Result:= l_revenu_simple* (1- l_taux_is)/ l_diviseur
                end;
              end; // with m_c_parameters_ref
            end; // _f_salaire

          procedure _itere_calcul(p_iteration: Integer; p_display: Boolean);
            var l_delta: Double;
                l_objectif_dividendes: Double;
                l_pourcentage_actuel: Double;
            begin
              with m_c_parameters_ref do
              begin
                if p_display= k_display_false
                  then push_stop_display(1);
                display_line;
                display('---');

                // -- revenu_net = revenu_avant_salaire- salaire - charges_salaire - interets- loyer- dividendes
                with m_c_calcul_charges_sarl_ref do
                begin
                  add_table_start('40_rev_IT', m_title);

                  compute_charges_sociales_sarl(m_salaire, m_cotisations_sociales_obligatoire_precedentes);

                  l_pourcent_charges:= m_pourcent_charges_salariales;
                end;

                m_dividendes:= 0;
                with m_c_calcul_is_ref do
                  display_debug(Format('exp %8.2f i %5.2f loy %8.2f sal %8.2f chge %8.2f div %8.2f  solde %8.2f',
                      [m_resultat_exploitation, m_interets_compte_associe, m_micro_foncier_brut,
                       m_salaire, m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl, m_dividendes,
                       m_resultat_exploitation- m_interets_compte_associe- m_micro_foncier_brut
                       - m_salaire- m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl- m_dividendes]));

                with m_c_calcul_is_ref do
                begin
                  add_table_start('46_rev_IT', m_title);


                  m_resultat_moins_salaires:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl;
                  m_benefice_avant_is:= m_resultat_moins_salaires;

                  calcule_is_interets_deductibles;

                  display(Format('ex %10.f  sal %10.2f chges %10.2f  sal+chge %10.2f',
                      [m_resultat_exploitation, m_salaire, m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl,
                       m_salaire+ m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl]));
                  display(Format('resu_moins_salaires %s %10.2f', [f_spaces(7), m_resultat_moins_salaires]));


                  calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_True);

                  l_objectif_dividendes:= m_salaire* (1- m_objectif_pourcent_salaire)/ m_objectif_pourcent_salaire;

                  l_delta:= l_objectif_dividendes- m_dividendes;

                  l_pourcentage_actuel:= m_salaire/ (m_salaire+  m_dividendes);
                  m_salaire:= m_salaire* (1- (l_pourcentage_actuel- m_objectif_pourcent_salaire) / 2);

                  display_debug(Format('  iter %2d  ben_net %8.2f obj_div %8.2f div_act %8.2f delta %10.2f  sal %8.2f  %%sal %5.2f',
                      [p_iteration, m_benefice_net, l_objectif_dividendes, m_dividendes, l_delta, m_salaire,
                       l_pourcentage_actuel  ]));
                end; // with m_c_calcul_is_ref
                if p_display= k_display_false
                  then pop_stop_display;
              end; // with m_c_parameters_ref
            end; // _itere_calcul

          var l_iteration: Integer;

          begin // _calcul_proportionnel
            with m_c_parameters_ref do
            begin
              l_pourcent_charges:= 0.60;
              l_taux_is:= 0.15;
              m_salaire:= _f_salaire(l_pourcent_charges);
              display_debug(Format('sal %10.2f', [m_salaire]));

              l_iteration:= 0;

              repeat
                _itere_calcul(l_iteration, k_display_false);

                Inc(l_iteration);
              until l_iteration> 5;

              _itere_calcul(l_iteration, k_display_true);
            end; // with m_c_parameters_ref
          end; // _calcul_proportionnel

        begin // _calcul_iteratif
          with m_c_parameters_ref do
          begin
            display_line;

            with m_c_calcul_is_ref do
              l_revenu_simple:= m_resultat_exploitation- m_micro_foncier_brut- m_interets_compte_associe;

            if m_objectif_pourcent_salaire= 0
              then _calcul_zero_salaire
              else
                if m_objectif_pourcent_salaire= 100
                  then _calcul_100_salaire
                  else _calcul_proportionnel;

            display_line;
            with m_c_calcul_is_ref do
            begin
              display_debug(Format('l_sal %8.2f l_div %8.2f ', [m_salaire, m_dividendes]));
              display_debug(Format('salaire+ dividendes %s %10.2f', [f_spaces(7), m_salaire+ m_dividendes]));
            end; // with m_c_calcul_is_ref

            with m_c_calcul_ir_ref do
            begin
              m_total_charges:= m_c_calcul_charges_sarl_ref.m_total_charges_salaire_sarl;

              m_revenu_imposable:= m_salaire;
              m_csg_deductible_ir:= m_c_calcul_charges_sarl_ref.m_csg_deductible_sarl;

              compute_ir;
            end; // with m_c_calcul_ir_ref
          end; // with m_c_parameters_ref
        end; // _calcul_iteratif

      var l_c_save_display_strings: tStrings;

      begin // compute_revenu_sarl
        with m_c_parameters_ref do
        begin
          l_c_save_display_strings:= s_c_strings;
          initialize_display(p_c_result_strings);
          clear_display;
          
          m_status_kind:= e_sarl_status;

          if (m_objectif_pourcent_salaire< 0) and (m_objectif_renvenu_net_societe= 0)
            then _calcul_litteral
            else _calcul_iteratif;

          initialize_display(l_c_save_display_strings);
        end; // with m_c_parameters_ref
      end; // compute_revenu_sarl

    procedure c_simulateur_revenus.compute_revenu_sas(p_c_result_strings: tStrings);

      procedure _calcul_litteral;
        begin
          with m_c_parameters_ref do
          begin
            with m_c_calcul_charges_sas_ref do
            begin
              add_table_start('10_rev', m_title);

              compute_charges_sociales_sas(m_salaire);
            end;

            with m_c_calcul_is_ref do
            begin
              add_table_start('20_rev', m_title);

              m_resultat_moins_salaires:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas;
              m_benefice_avant_is:= m_resultat_moins_salaires;

              calcule_is_interets_deductibles;

              display(f_display('resu_moins_salaires', m_resultat_moins_salaires));

              calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_false);
            end;

            with m_c_calcul_ir_ref do
            begin
              m_total_charges:= m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas;
              m_revenu_imposable:= m_salaire;
              m_csg_deductible_ir:= m_c_calcul_charges_sas_ref.m_csg_deductible_sas;

              compute_ir;
            end;
          end; // with m_c_parameters_ref
        end; // _calcul_litteral

      procedure _calcul_iteratif;
        const k_display_false= False;
              k_display_true= True;
        var l_revenu_simple: Double;

        procedure _calcul_zero_salaire;
          // var l_taux_is: Double;
          begin
            with m_c_parameters_ref do
            begin
              // -- pas de salaire

              with m_c_calcul_is_ref do
              begin
                add_table_start('46_rev_IT', m_title);

                m_resultat_moins_salaires:= m_resultat_exploitation;
                m_benefice_avant_is:= m_resultat_moins_salaires;

                calcule_is_interets_deductibles;

                display(Format('ex %10.f',
                    [m_resultat_exploitation]));
                display(Format('resu_moins_salaires %s %10.2f', [f_spaces(7), m_resultat_moins_salaires]));


                calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_True);

                display(Format('obj_ %8.2f', [m_dividendes]));
              end; // with m_c_calcul_is_ref

              m_salaire:= 0;
            end; // with m_c_parameters_ref
          end; // _calcul_zero_salaire

        procedure _display(p_text: String);
          begin
            if k_display_true
              then display(p_text);
          end; // _display

        procedure _calcul_100_salaire;
          var l_pourcent_charges: Double;

          procedure _itere_calcul(p_iteration: Integer; p_display: Boolean);
            var l_delta: Double; // , l_taux_is
            begin
              with m_c_parameters_ref do
              begin
                if p_display= k_display_false
                  then push_stop_display(1);

                _display('');
                _display('---');
                with m_c_calcul_charges_sas_ref do
                begin
                  add_table_start('40_rev_IT', m_title);

                  compute_charges_sociales_sas(m_salaire);

                  l_pourcent_charges:= m_pourcent_charges_salariales;
                end;

                with m_c_calcul_is_ref do
                  display_debug(Format('exp %8.2f i %5.2f loy %8.2f sal %8.2f chge %8.2f solde %8.2f',
                      [m_resultat_exploitation, m_interets_compte_associe, m_micro_foncier_brut,
                       m_salaire, m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas,
                       m_resultat_exploitation- m_interets_compte_associe- m_micro_foncier_brut
                       - m_salaire- m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas]));

                with m_c_calcul_is_ref do
                begin
                  add_table_start('46_rev_IT', m_title);

                  m_resultat_moins_salaires:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas;
                  m_benefice_avant_is:= m_resultat_moins_salaires;

                  calcule_is_interets_deductibles;

                  display(Format('ex %10.f  sal %10.2f chges %10.2f  sal+chge %10.2f',
                      [m_resultat_exploitation, m_salaire, m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas,
                       m_salaire+ m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas]));
                  display(Format('resu_moins_salaires %s %10.2f', [f_spaces(7), m_resultat_moins_salaires]));

                  calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_true);

                  l_delta:= m_benefice_net/ (1- m_taux_is_moyen)/ (1 + m_c_calcul_charges_sas_ref.m_pourcent_charges_salariales);

                  m_salaire:= m_salaire+ l_delta;

                  display_debug(Format('  iter %2d  ben_net %8.2f pourc_charges %5.2f taux_is %5.2f delta %5.2f sal %8.2f',
                      [p_iteration, m_benefice_net,
                       m_c_calcul_charges_sas_ref.m_pourcent_charges_salariales, m_taux_is_moyen, m_salaire]));
                end; // with m_c_calcul_is_ref

                if p_display= k_display_false
                  then pop_stop_display;
              end; // with m_c_parameters_ref
            end; // _itere_calcul

          var l_iteration: Integer;

          begin // _calcul_100_salaire
            with m_c_parameters_ref do
            begin
              l_pourcent_charges:= 0.60;
              m_salaire:= l_revenu_simple/ (1+ l_pourcent_charges);
              l_iteration:= 0;

              repeat
                _itere_calcul(l_iteration, k_display_false);

                Inc(l_iteration);
              until l_iteration> 5;

              _itere_calcul(l_iteration, k_display_true);

              m_dividendes:= 0;
            end; // with m_c_parameters_ref
          end; // _calcul_100_salaire

        procedure _calcul_proportionnel;
          var l_pourcent_charges, l_taux_is: Double;

          function _f_salaire(p_pourcent_charges: Double): Double;
            var l_diviseur: Double;
            begin
              with m_c_parameters_ref do
              begin
                with m_c_calcul_is_ref do
                begin
                  l_diviseur:= (1- m_objectif_pourcent_salaire)/ m_objectif_pourcent_salaire
                      + (1+ l_pourcent_charges)* (1- l_taux_is);

                  Result:= l_revenu_simple* (1- l_taux_is)/ l_diviseur
                end;
              end; // with m_c_parameters_ref
            end; // _f_salaire

          procedure _itere_calcul(p_iteration: Integer; p_display: Boolean);
            var l_delta: Double;
                l_objectif_dividendes: Double;
                l_pourcentage_actuel: Double;
            begin
              with m_c_parameters_ref do
              begin
                if p_display= k_display_false
                  then push_stop_display(1);
                display_line;
                display('---');

                // -- revenu_net = revenu_avant_salaire- salaire - charges_salaire - interets- loyer- dividendes
                // -- %_salaire = salaires / (salaire+ dividendes)
                with m_c_calcul_charges_sas_ref do
                begin
                  add_table_start('40_rev_IT', 'SAS_sss');

                  compute_charges_sociales_sas(m_salaire);

                  l_pourcent_charges:= m_pourcent_charges_salariales;
                end;

                m_dividendes:= 0;
                with m_c_calcul_is_ref do
                  display_debug(Format('exp %8.2f i %5.2f loy %8.2f sal %8.2f chge %8.2f div %8.2f  solde %8.2f',
                      [m_resultat_exploitation, m_interets_compte_associe, m_micro_foncier_brut,
                       m_salaire, m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas, m_dividendes,
                       m_resultat_exploitation- m_interets_compte_associe- m_micro_foncier_brut
                       - m_salaire- m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas- m_dividendes]));

                with m_c_calcul_is_ref do
                begin
                  add_table_start('46_rev_IT', m_title);

                  m_resultat_moins_salaires:= m_resultat_exploitation- m_salaire- m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas;
                  m_benefice_avant_is:= m_resultat_moins_salaires;

                  calcule_is_interets_deductibles;

                  display(Format('ex %10.f  sal %10.2f chges %10.2f  sal+chge %10.2f',
                      [m_resultat_exploitation, m_salaire, m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas,
                       m_salaire+ m_c_calcul_charges_sas_ref.m_total_charges_salaire_sas]));
                  display(Format('resu_moins_salaires %s %10.2f', [f_spaces(7), m_resultat_moins_salaires]));

                  calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_true);

                  l_objectif_dividendes:= m_salaire* (1- m_objectif_pourcent_salaire)/ m_objectif_pourcent_salaire;
                  m_dividendes:= m_benefice_net;
                  // m_dividendes:= m_dividendes;
                  m_reserve:= 0;

                  l_delta:= l_objectif_dividendes- m_dividendes;

                  l_pourcentage_actuel:= m_salaire/ (m_salaire+  m_dividendes);
                  m_salaire:= m_salaire* (1- (l_pourcentage_actuel- m_objectif_pourcent_salaire) / 2);

                  display_debug(Format('  iter %2d  ben_net %8.2f obj_div %8.2f div_ect %8.2f delta %10.2f  sal %8.2f  %%sal %5.2f',
                      [p_iteration, m_benefice_net, l_objectif_dividendes, m_dividendes, l_delta, m_salaire,
                       l_pourcentage_actuel  ]));
                end; // with m_c_calcul_is_ref
                if p_display= k_display_false
                  then pop_stop_display;
              end; // with m_c_parameters_ref
            end; // _itere_calcul

          var l_iteration: Integer;

          begin // _calcul_proportionnel
            with m_c_parameters_ref do
            begin
              l_pourcent_charges:= 0.60;
              l_taux_is:= 0.15;
              m_salaire:= _f_salaire(l_pourcent_charges);
              display_debug(Format('sal %10.2f', [m_salaire]));

              l_iteration:= 0;

              repeat
                _itere_calcul(l_iteration, k_display_false);

                Inc(l_iteration);
              until l_iteration> 5;

              _itere_calcul(l_iteration, k_display_true);
            end; // with m_c_parameters_ref
          end; // _calcul_proportionnel

        begin // _calcul_iteratif
          with m_c_parameters_ref do
          begin
            display_line;

            with m_c_calcul_is_ref do
              l_revenu_simple:= m_resultat_exploitation- m_micro_foncier_brut- m_interets_compte_associe;

            if m_objectif_pourcent_salaire= 0
              then _calcul_zero_salaire
              else
                if m_objectif_pourcent_salaire= 100
                  then _calcul_100_salaire
                  else _calcul_proportionnel;

            display_line;
            with m_c_calcul_is_ref do
            begin
              display_debug(Format('l_sal %8.2f l_div %8.2f ', [m_salaire, m_dividendes]));
              display_debug(Format('salaire+ dividendes %s %10.2f', [f_spaces(7), m_salaire+ m_dividendes]));
            end; // with m_c_calcul_is_ref

            with m_c_calcul_ir_ref do
            begin
              m_revenu_imposable:= m_salaire;
              m_csg_deductible_ir:= m_c_calcul_charges_sas_ref.m_csg_deductible_sas;

              compute_ir;
            end; // with m_c_calcul_ir_ref
          end; // with m_c_parameters_ref
        end; // _calcul_iteratif

      var l_c_save_display_strings: tStrings;

      begin // compute_revenu_sas
        l_c_save_display_strings:= s_c_strings;
        initialize_display(p_c_result_strings);
        clear_display;

        with m_c_parameters_ref do
        begin
          m_status_kind:= e_sas_status;

          if (m_objectif_pourcent_salaire< 0) and (m_objectif_renvenu_net_societe= 0)
            then _calcul_litteral
            else _calcul_iteratif;
        end; // with m_c_parameters_ref

        initialize_display(l_c_save_display_strings);
      end; // compute_revenu_sas

    procedure c_simulateur_revenus.compute_revenu_sas_IR(p_c_result_strings: tStrings);

      procedure _calcule_revenu_sas_0_IR;
        // -- use the value set by the user
        begin
          with m_c_parameters_ref do
          begin
            with m_c_calcul_is_ref do
            begin
              m_revenu_sas_IR:= 0;

              m_benefice_avant_is:= m_resultat_exploitation;

              calcule_is_interets_deductibles;

              calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_false);

              display(f_display(' - dividendes', m_dividendes));
              m_revenu_sas_IR:= m_benefice_net- m_dividendes;
              display(f_display(' = reste_sas', m_benefice_net- m_dividendes));

              m_salaire:= m_revenu_sas_IR;
            end; // with m_c_calcul_is_ref
          end; // with m_c_parameters_ref
        end; // _calcule_revenu_sas_0_IR

      procedure _calcule_revenu_sas_IR_dividendes;
        // -- calcul pour verser un pourcentage de salaire donn
        begin
          with m_c_parameters_ref do
          begin
            with m_c_calcul_is_ref do
            begin
              display(f_display('revenu_exploitation ', m_resultat_exploitation));

              m_benefice_avant_is:= m_resultat_exploitation;

              calcule_is_interets_deductibles;
              calcule_is_sur_revenu_imposable(k_set_dividendes_to_benefice_net_true);

              if m_objectif_pourcent_salaire= 0
                then m_revenu_sas_IR:= 0
                else m_revenu_sas_IR:= m_benefice_net/ (1 + (1- m_objectif_pourcent_salaire)/ m_objectif_pourcent_salaire);
              m_dividendes:= m_benefice_net- m_revenu_sas_IR;
              m_salaire:= m_revenu_sas_IR;

              display(f_display(' - dividendes', m_dividendes));
              display(f_display(' - revenu_dir_IR', m_revenu_sas_IR));
              display(f_display(' = reste_sas', m_benefice_net- m_revenu_sas_IR- m_dividendes));
            end; // with m_c_calcul_is_ref
          end; // with m_c_parameters_ref
        end; // _calcule_revenu_sas_IR_dividendes

      var l_c_save_display_strings: tStrings;

      begin // compute_revenu_sas_IR
        l_c_save_display_strings:= s_c_strings;
        initialize_display(p_c_result_strings);
        clear_display;

        with m_c_parameters_ref do
        begin
          with m_c_calcul_is_ref do
          begin
            add_table_start('50_rev', m_title);

            calcule_is_interets_deductibles;

            m_status_kind:= e_sas_ir_status;
            if (m_objectif_pourcent_salaire< 0) and (m_objectif_renvenu_net_societe= 0)
              then _calcule_revenu_sas_0_IR
              else _calcule_revenu_sas_IR_dividendes;
            m_reserve:= 0;

            display_debug(Format('=== %s rev_IR %10.2f div %10.2f rev_IR+ dividendes %10.2f',
                 [f_spaces(7), m_revenu_sas_IR, m_dividendes, m_revenu_sas_IR+ m_dividendes]));
          end; // with m_c_calcul_is_ref

          with m_c_calcul_ir_ref do
          begin
            m_revenu_imposable:= m_revenu_sas_IR;
            m_csg_deductible_ir:= 0;

            m_revenu_imposable:= m_revenu_sas_IR;

            compute_ir;
          end; // with m_c_calcul_ir_ref
        end; // with m_c_parameters_ref

        initialize_display(l_c_save_display_strings);
      end; // compute_revenu_sas_IR

    Destructor c_simulateur_revenus.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_simulateur_revenus
    end. // u_c_simulateur_revenus


